import asyncio, json, csv, httpx
import websockets.exceptions
from datetime import datetime
from idun_data_models import GuardianRecording, json_pydantic
from idun_data_models.api_data_model import DataStreams
from idun_tools import logger
from ..data_generator import generate_queue, GeneratorTime


class MockGuardianDataGenerator:
    def __init__(self):
        self.message_queue = asyncio.Queue(maxsize=int(1e7))
        self.recordingID_queue = asyncio.Queue(maxsize=10)
        self.recordingIDs = []
        self.stop = asyncio.Event()

    async def generate(
        self,
        devices: int,
        connections_per_device: int,
        p_connection_end: float,
        generator_time_type: GeneratorTime,
        delay: float = 0,
        mock_recorded: bool = False,
        random_seed: int = 0,
        stop: asyncio.Event = None,
        finish_generating: asyncio.Event = asyncio.Event(),
        enableStreamer: bool = True,
        first_device=0,
    ):
        if stop is None:
            stop = self.stop
        return await generate_queue(
            logger,
            self.message_queue,
            self.recordingID_queue,
            devices,
            connections_per_device,
            p_connection_end,
            generator_time_type,
            delay,
            mock_recorded,
            random_seed,
            stop,
            finish_generating,
            enableStreamer,
            first_device,
        )

    async def gather_recordingIDs(self) -> list[dict]:
        if len(self.recordingIDs) == 0:
            self.recordingIDs = await self.recordingID_queue.get()
        return self.recordingIDs

    async def register_recordings(self, rest_api_url: str) -> None:
        async def register_recording(client, deviceID, recordingID):
            payload = (
                json_pydantic(
                    GuardianRecording(
                        recordingID,
                        deviceID,
                        "mock data generator",
                        DataStreams(bandpass_eeg=True),
                    )
                ),
            )
            # TODO: fix json_pydantic
            if isinstance(payload, tuple):
                payload = payload[0]
            logger.bind(payload=payload).info("Registering recording")
            await http_post(
                client,
                f"{rest_api_url}/api/recording/{deviceID}/recordings",
                auth=(deviceID, ""),
                content=payload,
            )

        recordingIDs = await self.gather_recordingIDs()
        async with httpx.AsyncClient() as client, asyncio.TaskGroup() as tg:
            for i in recordingIDs:
                tg.create_task(
                    register_recording(client, i["deviceID"], i["recordingID"])
                )
        logger.info("Recordings registered")

    async def stop_recordings(self, rest_api_url: str) -> None:
        async def stop_recording(client, deviceID, recordingID):
            await http_post(
                client,
                f"{rest_api_url}/api/recording/{deviceID}/{recordingID}/stop",
                auth=(deviceID, ""),
            )

        recordingIDs = await self.gather_recordingIDs()
        async with httpx.AsyncClient() as client, asyncio.TaskGroup() as tg:
            for i in recordingIDs:
                tg.create_task(stop_recording(client, i["deviceID"], i["recordingID"]))
        logger.info("Recordings stopped")

    async def send(self, websocket, stop: asyncio.Event = None):
        "Given a connected Websocket, `send` will directly put data in it from the queue."

        if stop is None:
            stop = self.stop

        async def may_block():
            message = await self.message_queue.get()
            try:
                await websocket.send(json_pydantic(message))
            except websockets.exceptions.ConnectionClosedError:
                await self.message_queue.put(message)
                raise

        try:
            while not stop.is_set():
                await asyncio.wait_for(may_block(), timeout=6.0)
        except asyncio.TimeoutError:
            if self.message_queue.empty():
                logger.info("Sending timed out. Queue empty; stopping")
                stop.set()
            else:
                logger.info(
                    f"Sending timed out. Messages in queue: {self.message_queue.qsize()}; reconnecting to websocket"
                )
                raise

    async def log_stream(self, websocket, stop: asyncio.Event = None):
        "Given a connected Websocket, `read_stream` will read data from it and log it."

        if stop is None:
            stop = self.stop

        async def may_block(latency_f):
            message = json.loads(await websocket.recv())
            if "bp_filter_eeg" in message:
                ts = message["bp_filter_eeg"]["timestamp"][0]
                latency = datetime.now().timestamp() - ts
                samples_in_message = len(message["bp_filter_eeg"]["timestamp"])
                logger.info(
                    f"Latency: {latency} ({samples_in_message} samples in message)"
                )
                latency_f.writerow(
                    {
                        "timestamp": ts,
                        "deviceID": message["deviceID"],
                        "latency": latency,
                    }
                )

        try:
            with open("latency.log", "a") as latency_f:
                latency_csv = csv.DictWriter(
                    latency_f,
                    dialect="excel",
                    fieldnames=["timestamp", "deviceID", "latency"],
                )
                if latency_f.tell() == 0:
                    latency_csv.writeheader()
                while not stop.is_set():
                    await asyncio.wait_for(may_block(latency_csv), timeout=6.0)
        except asyncio.TimeoutError:
            logger.info("Receiving timed out")

    async def send_kinesis(
        self, kinesis_client, stream_name, stop: asyncio.Event = None
    ):
        "Put data to Kinesis from the queue"

        if stop is None:
            stop = self.stop

        async def may_block():
            message = await self.message_queue.get()
            kinesis_client.put_record(
                StreamName=stream_name,
                Data=json.dumps(apiGatewayTransform(message)),
                PartitionKey=message["deviceID"],
            )

        try:
            while not stop.is_set():
                await asyncio.wait_for(may_block(), timeout=6.0)
        except asyncio.TimeoutError:
            logger.info("Sending timed out, stopping.")
            stop.set()


async def http_post(client: httpx.AsyncClient, url: str, auth, content=None):
    try:
        response = await client.post(
            url,
            content=content,
            auth=auth,
            headers=httpx.Headers({"Content-Type": "application/json"}),
        )
        response.raise_for_status()
    except httpx.HTTPStatusError as err:
        status_code = err.response.status_code
        logger.bind(
            request=err.request,
            response=err.response,
            reason=err.response.reason_phrase,
            reason_text=err.response.text,
            status_code=status_code,
            http_version=err.response.http_version,
            http_headers=err.response.headers,
            elapsed_sec=err.response.elapsed.total_seconds(),
        ).warning(f"Received error code from POST {url}")
    except httpx.RequestError as err:
        logger.bind(error=err).error("HTTP request error")


def apiGatewayTransform(message: dict):
    "Wrap message with the API gateway's template transform"
    message["connectionID"] = "a"
    return message



import threading, random, time, asyncio
from dataclasses import asdict
from idun_data_models import Message
from idun_tools import logger
from .connection_pool import DeviceConnectionPool, GeneratorTime


async def generate_queue(
    logger,
    message_queue: asyncio.Queue,
    recordingID_queue: asyncio.Queue,
    devices: int,
    connections_per_device: int,
    p_connection_end: float,
    generator_time_type: GeneratorTime,
    delay: float = 0,
    mock_recorded: bool = False,
    random_seed: int = 0,
    stop: asyncio.Event = asyncio.Event(),
    finish_generating: asyncio.Event = asyncio.Event(),
    enableStreamer: bool = True,
    first_device=0,
) -> set[Message]:
    """
    Generate random device messages and put them on the `message_queue`.
    When the connection pool finishes or a stop signal arrives, it terminates, and returns a set of all messages.

    Parameters:
    - `message_queue`: queue where messages will be put
    - `recordingID_queue`: only 1 item will be put in the queue: a list of the generated recordingIDs
    """
    connectionPool = DeviceConnectionPool(
        devices=devices,
        connections_per_device=connections_per_device,
        p_connection_end=p_connection_end,
        p_impedance_measurement=0,
        mock_recorded=mock_recorded,
        random_seed=random_seed,
        generator_time_type=generator_time_type,
        enableStreamer=enableStreamer,
        first_device=first_device,
    )
    logger.info("> Generating test data")
    logger.info(
        f"> Recording and Device ID for all recordings that should be produced after running this generator to completion: {connectionPool.allRecordingIDs()}"
    )
    await recordingID_queue.put(connectionPool.allRecordingIDs())
    messageSet = set()
    while not stop.is_set() and not connectionPool.done():
        data = connectionPool.next_message()
        if data:
            try:
                await asyncio.wait_for(message_queue.put(asdict(data)), timeout=6.0)
                messageSet.add(data)
                await asyncio.sleep(delay / devices)
            except asyncio.TimeoutError:
                if message_queue.full():
                    logger.error("Generator queue full! Stopping generation")
                else:
                    logger.error("Generator timeout! Stopping generation")
                stop.set()
    finish_generating.set()
    logger.info("> Stop generating test data")
    return messageSet


def generate_kinesis(
    kinesis_client,
    devices: int,
    connections_per_device: int,
    p_connection_end: float,
    p_impedance_measurement: float = 0,
    mock_recorded: bool = False,
    random_seed: int = 0,
    stop: threading.Event = threading.Event(),
) -> set[Message]:
    """
    Generate random device messages and put them on the Kinesis stream.
    When the connection pool finishes or a stop signal arrives, it terminates, and returns a set of all messages.
    """
    logger.info("> Generating test data")
    connectionPool = DeviceConnectionPool(
        devices=devices,
        connections_per_device=connections_per_device,
        p_connection_end=p_connection_end,
        p_impedance_measurement=p_impedance_measurement,
        mock_recorded=mock_recorded,
        random_seed=random_seed,
    )
    messageSet = set()
    while not stop.is_set() and not connectionPool.done():
        data = connectionPool.next_message()
        if data:
            kinesis_client.put_record(
                Data=asdict(data),
                PartitionKey=data.deviceID,
            )
            messageSet.add(data)
    logger.info("> Stop generating test data")
    return messageSet


def randomDecodedPayload():
    "Return a random, decoded payload. Mock the decoding function."
    return {
        "eeg_ch1": list(
            map(
                lambda x: {
                    "timestamp": time.time(),
                    "value": random.random(),
                },
                range(0, 20),
            )
        ),
        "acc": list(
            map(
                lambda x: {
                    "timestamp": time.time(),
                    "x": random.random(),
                    "y": random.random(),
                    "z": random.random(),
                },
                range(0, 3),
            )
        ),
        "magn": list(
            map(
                lambda x: {
                    "timestamp": time.time(),
                    "x": random.random(),
                    "y": random.random(),
                    "z": random.random(),
                },
                range(0, 3),
            )
        ),
        "gyro": list(
            map(
                lambda x: {
                    "timestamp": time.time(),
                    "x": random.random(),
                    "y": random.random(),
                    "z": random.random(),
                },
                range(0, 3),
            )
        ),
    }



import asyncio, argparse, boto3, os
import websockets.client, websockets.exceptions
from idun_tools import logger
from ..data_generator import GeneratorTime
from . import MockGuardianDataGenerator


generator = MockGuardianDataGenerator()


async def generate_send(
    websocket_url,
    rest_api_url,
    devices=3,
    connections_per_device=1,
    p_connection_end=1e-3,
    delay=0,
    enableStreamer=True,
):
    if delay > 0:
        GENERATOR_TIME_TYPE = GeneratorTime.IDEAL_DEVICE
    else:
        GENERATOR_TIME_TYPE = GeneratorTime.CLIENT
    async with asyncio.TaskGroup() as tg:
        tasks = {}
        tasks["generator"] = tg.create_task(
            generator.generate(
                devices,
                connections_per_device,
                p_connection_end,
                GENERATOR_TIME_TYPE,
                delay=delay,
                enableStreamer=enableStreamer,
                first_device=1,
            )
        )
        # POST the recording at the REST endpoint
        await generator.register_recordings(rest_api_url)

        # Connect to the websocket and send data
        try:
            async for websocket in websockets.client.connect(websocket_url):
                reconnect = False
                try:
                    async with asyncio.TaskGroup() as websocket_tg:
                        websocket_tg.create_task(generator.send(websocket))
                        websocket_tg.create_task(generator.log_stream(websocket))
                except* (
                    websockets.exceptions.ConnectionClosed,
                    websockets.exceptions.ConnectionClosedError,
                ) as err:
                    # This also captures ConnectionClosedOK. We still want to reconnect even if the cloud legitimately closes the connection.
                    logger.bind(error=err).info(
                        "Websocket disconnected; reconnecting..."
                    )
                    reconnect = True
                except* (asyncio.CancelledError, asyncio.TimeoutError):
                    logger.info("Websocket timeout; reconnecting...")
                    reconnect = True

                if not reconnect:
                    break
        except* Exception as err:
            logger.bind(error=err).info("Websocket error")
        finally:
            generator.stop.set()
            await generator.stop_recordings(rest_api_url)


async def generate_send_kinesis(
    kinesis_stream_name,
    aws_endpoint_url,
    devices=3,
    connections_per_device=1,
    p_connection_end=1e-3,
    delay=0,
    enableStreamer=True,
):
    if delay > 0:
        GENERATOR_TIME_TYPE = GeneratorTime.IDEAL_DEVICE
    else:
        GENERATOR_TIME_TYPE = GeneratorTime.CLIENT
    await asyncio.gather(
        generator.generate(
            devices,
            connections_per_device,
            p_connection_end,
            GENERATOR_TIME_TYPE,
            delay=delay,
            enableStreamer=enableStreamer,
        ),
        generator.send_kinesis(
            boto3.client("kinesis", endpoint_url=aws_endpoint_url), kinesis_stream_name
        ),
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="MockDeviceDataGenerator",
        description="Send mocked device data to the cloud (for tests)",
    )
    data_endpoint = parser.add_mutually_exclusive_group(required=True)
    data_endpoint.add_argument(
        "-w", "--websocket", help="Websocket URL where data will be sent"
    )
    data_endpoint.add_argument(
        "-k",
        "--kinesis_stream_name",
        help="Kinesis stream name where data will be sent; specify together with `--aws_endpoint_url`",
    )
    parser.add_argument(
        "-r",
        "--rest_api_url",
        required=True,
        help="URL of the REST API where recordings are registered",
    )
    parser.add_argument(
        "--aws_endpoint_url",
        help="AWS endpoint URL; specify together with `--kinesis_stream_name`",
    )
    parser.add_argument(
        "--avg_duration",
        default=1e3,
        type=float,
        help="Average length of recording in messages",
    )
    parser.add_argument(
        "-d",
        "--devices",
        default=5,
        type=int,
        help="Number of devices streaming simultaneously",
    )
    parser.add_argument(
        "--delay_sec",
        default=0,
        type=float,
        help="Add a delay between generated messages per device. Delay of a real device is 0.080",
    )
    parser.add_argument(
        "--disable_live_stream",
        action="store_true",
        help="Toggle the live streaming of samples",
    )
    args = parser.parse_args()

    try:
        os.remove("latency.log")
    except FileNotFoundError:
        pass

    # NOTE: Connections per device can't be > 1 because we register all the recordings together in the beginning
    if args.websocket:
        asyncio.run(
            generate_send(
                args.websocket,
                args.rest_api_url,
                p_connection_end=1 / args.avg_duration,
                devices=args.devices,
                delay=args.delay_sec,
                enableStreamer=not args.disable_live_stream,
            )
        )
    elif args.kinesis_stream_name:
        asyncio.run(
            generate_send_kinesis(
                args.kinesis_stream_name,
                args.aws_endpoint_url,
                p_connection_end=1 / args.avg_duration,
                devices=args.devices,
                delay=args.delay_sec,
                enableStreamer=not args.disable_live_stream,
            )
        )



import random, datetime, secrets
from enum import Enum
from pathlib import Path
from uuid import uuid4
from idun_data_models import Message, DevicePacket


class GeneratorTime(Enum):
    IDEAL_DEVICE = 1
    "Simulates an ideal device with perfect clock that sends 1 data packet every 80ms"
    CLIENT = 2
    "Uses the client's local time for each message's timestamp"


class DeviceConnectionPool:
    """
    DeviceConnectionPool generates device and connection IDs for mock messages.
    It simulates streams of messages arriving simultaneously from a given number of devices,
    which stop after a random number of messages (geometric distribution).
    The last message for each device has a `stop` field set to true.
    The next time the device is picked, it will start a new connection.
    The device payload isn't encrypted.
    """

    def __init__(
        self,
        devices: int,
        connections_per_device: int,
        p_connection_end: float,
        p_impedance_measurement: float,
        mock_recorded: bool,
        random_seed: int = 0,
        generator_time_type: GeneratorTime = GeneratorTime.IDEAL_DEVICE,
        enableStreamer=True,
        first_device=0,
    ) -> None:
        start_time = datetime.datetime.now()
        self.devices = [str(i) for i in range(first_device, first_device + devices)]
        self.openDeviceIdx = list(range(0, len(self.devices)))
        self.device_time = [start_time for i in self.openDeviceIdx]
        self.connections_per_device = connections_per_device
        self.p_connection_end = p_connection_end
        self.generator_time_type = generator_time_type
        self.p_impedance_measurement = p_impedance_measurement
        self.mock_recorded = mock_recorded
        self.enableStreamer = enableStreamer
        self.connections = {d: 0 for d in self.openDeviceIdx}
        self.first_message = {d: True for d in self.openDeviceIdx}
        self.random = random.Random()
        self.random.seed(random_seed)
        self.connection_prefix = uuid4()

    def next_message(self) -> Message | None:
        device_idx = self.random.choice(self.openDeviceIdx)
        deviceID = str(self.devices[device_idx])
        connection = self.connections[device_idx]
        if connection == self.connections_per_device:
            self.openDeviceIdx.remove(device_idx)
            return None
        # Never end the connection at the first message
        connectionEnd = (
            not self.first_message[device_idx]
            and self.random.random() < self.p_connection_end
        )
        self.first_message[device_idx] = False

        # Advance the device time by the real duration of a device packet; ie 80ms
        self.device_time[device_idx] = self.device_time[
            device_idx
        ] + datetime.timedelta(milliseconds=80)
        if self.generator_time_type == GeneratorTime.IDEAL_DEVICE:
            message_timestamp = str(self.device_time[device_idx])
        elif self.generator_time_type == GeneratorTime.CLIENT:
            message_timestamp = str(datetime.datetime.now())
        else:
            # Is there another case?
            message_timestamp = str(datetime.datetime.now())

        # Advance state
        if connectionEnd:
            self.connections[device_idx] += 1
            self.first_message[device_idx] = True
        # With a small probability, emit an impedance measurement. Otherwise, emit a normal device message.
        if self.random.random() < self.p_impedance_measurement:
            # Send an impedance message
            return Message(
                deviceTimestamp=message_timestamp,
                deviceID=deviceID,
                recordingID=connectionUUID(
                    self.connection_prefix, connection, deviceID
                ),
                connectionID=deviceID,
                stop=connectionEnd,
                recorded=self.mock_recorded and connectionEnd,
                payload=None,
                impedance=self.random.random(),
                enableStreamer=self.enableStreamer,
            )
        else:
            return Message(
                deviceTimestamp=message_timestamp,
                deviceID=deviceID,
                recordingID=connectionUUID(
                    self.connection_prefix, connection, deviceID
                ),
                connectionID=str(device_idx),
                stop=connectionEnd,
                recorded=self.mock_recorded and connectionEnd,
                payload=random_payload(deviceID),
                enableStreamer=self.enableStreamer,
            )

    def allRecordingIDs(self) -> list:
        return [
            {
                "deviceID": device,
                "recordingID": connectionUUID(
                    self.connection_prefix, connection, device
                ),
            }
            for device in self.devices
            for connection in range(0, self.connections_per_device)
        ]

    def done(self) -> bool:
        return all(
            [c >= self.connections_per_device for c in self.connections.values()]
        )


def connectionUUID(prefix, recordingID, deviceID) -> str:
    # NOTE: this should be a UUID, but for the moment a concatenation will do
    return str(prefix) + str(recordingID) + str(deviceID)


def random_device_packet() -> DevicePacket:
    """
    Generate mock data for GDK 2.0, with mock data of two channels and counter
    """

    def randomSampleList(nSamples=20):
        return random.sample(range(0, 256), nSamples)

    def samplePointList(*args):
        return list(map(list, zip(*args)))

    return DevicePacket(
        ch1=randomSampleList(),
        ch2=randomSampleList(),
        acc=samplePointList(randomSampleList(), randomSampleList(), randomSampleList()),
        magn=samplePointList(
            randomSampleList(), randomSampleList(), randomSampleList()
        ),
        gyro=samplePointList(
            randomSampleList(), randomSampleList(), randomSampleList()
        ),
    )


def static_recordingKey(deviceID=None) -> bytes:
    "Decryption key to be used with the `random_payload` function."
    keys = {
        None: "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f",  # old static key
        "0": "7e89fbec9fa8cf257a04ad4fe603f57a85adec2598b6d3f98229c02c3ecad00c",
    }
    if deviceID not in keys:
        key = keys[None]
    else:
        key = keys[deviceID]
    return bytes.fromhex(key)


def random_recordingKey() -> bytes:
    "Dynamically generated decryption key"
    return secrets.token_bytes(32)


def random_payload(deviceID=None) -> str:
    """
    Pick a random base64 encoded IGEB data packet from a short list of test samples.

    Example:

    Take a random device payload `p` and convert it into a dictionary that looks like a `DevicePacket`.
    ```
    from idun_guardian_client.igeb_utils_internal import decrypt_data, convert_igeb_data
    p= random_payload()
    d= convert_igeb_data(decrypt_data(base64.b64decode(p)[2:-1]))
    ```

    ! Bug: We can't use `dacite.from_dict` to create a DevicePacket:
    ```
    dacite.exceptions.WrongTypeError: wrong value type for field "acc" - should be "list" instead of value
    "[[-1.014, 2.337, -9.746], [-1.014, 2.337, -9.746], [-1.014, 2.337, -9.746], [-1.014, 2.337, -9.746], [-1.014, 2.337, -9.746], [-1.014, 2.337, -9.746], [-1.014, 2.337, -9.746], [-1.014, 2.337, -9.746], [-1.014, 2.337, -9.746], [-1.014, 2.337, -9.746], [-0.9380000000000001, 2.3770000000000002, -9.693], [-0.9380000000000001, 2.3770000000000002, -9.693], [-0.9380000000000001, 2.3770000000000002, -9.693], [-0.9380000000000001, 2.3770000000000002, -9.693], [-0.9380000000000001, 2.3770000000000002, -9.693], [-0.9380000000000001, 2.3770000000000002, -9.693], [-0.9380000000000001, 2.3770000000000002, -9.693], [-0.9380000000000001, 2.3770000000000002, -9.693], [-0.9380000000000001, 2.3770000000000002, -9.693], [-0.9380000000000001, 2.3770000000000002, -9.693]]"
    of type "list"
    ```
    """
    test_data_path = (
        Path(__file__).parent / f"igeb_test_packets-{deviceID}.txt"
    ).resolve()
    if not test_data_path.is_file():
        test_data_path = (Path(__file__).parent / "igeb_test_packets.txt").resolve()
    with open(test_data_path, "rt", encoding="ascii") as testdata_f:
        lines = testdata_f.readlines()
    return random.sample(lines, 1)[0].strip()
